""" Unit tests for the problem interface."""

import unittest
import os

from tempfile import mkdtemp
from collections import OrderedDict

from cosapp.tools.problem_viewer.problem_viewer import _get_viewer_data, view_model
from cosapp.tests.library.systems.others import ComplexTurbofan


class TestViewModelData(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.dir = mkdtemp()
        self.problem_filename = os.path.join(self.dir, "problem_n2")
        self.problem_html_filename = self.problem_filename
        self.expected_tree = OrderedDict(
            [
                ("name", "turbofan"),
                ("type", "subsystem"),
                ("subsystem_type", "group"),
                (
                    "children",
                    [
                        OrderedDict(
                            [
                                ("name", "atm"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "inlet"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "W_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "fanC"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "gh"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "mech_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "XN"),                                                                
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "PW"),                                                                
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "ductC"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "merger"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "duct"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "A",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "cst_loss",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "glp",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "outwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "PR",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        )
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "bleed"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "split_ratio",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        )
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fan"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "inwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "gh",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        )
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "mech_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "XN",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "PW",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "outwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "pcnr",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "pr",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "effis",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "wr",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "PWfan",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "merger"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl1_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "duct"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "A"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "cst_loss"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "glp"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "PR"),
                                                                ("type", "unknown"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "bleed"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "split_ratio"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl1_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "noz"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Acol"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Aexit"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "WRnozzle"),
                                                                ("type", "unknown"),
                                                            ]
                                                        )
                                                    ],
                                                ),

                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                    ],
                ),
            ]
        )

        self.expected_conns = [
            OrderedDict([("src", "atm.fl_out.Tt"), ("tgt", "inlet.fl_in.Tt")]),
            OrderedDict([("src", "atm.fl_out.Pt"), ("tgt", "inlet.fl_in.Pt")]),
            OrderedDict([("src", "inlet.fl_out.Tt"), ("tgt", "fanC.fl_in.Tt")]),
            OrderedDict([("src", "inlet.fl_out.Pt"), ("tgt", "fanC.fl_in.Pt")]),
            OrderedDict([("src", "inlet.fl_out.W"), ("tgt", "fanC.fl_in.W")]),
            OrderedDict([("src", "fanC.fl_out.Tt"), ("tgt", "merger.fl1_in.Tt")]),
            OrderedDict([("src", "fanC.fl_out.Pt"), ("tgt", "merger.fl1_in.Pt")]),
            OrderedDict([("src", "fanC.fl_out.W"), ("tgt", "merger.fl1_in.W")]),
            OrderedDict([("src", "bleed.fl2_out.Tt"), ("tgt", "merger.fl2_in.Tt")]),
            OrderedDict([("src", "bleed.fl2_out.Pt"), ("tgt", "merger.fl2_in.Pt")]),
            OrderedDict([("src", "bleed.fl2_out.W"), ("tgt", "merger.fl2_in.W")]),
            OrderedDict([("src", "merger.fl_out.Tt"), ("tgt", "duct.fl_in.Tt")]),
            OrderedDict([("src", "merger.fl_out.Pt"), ("tgt", "duct.fl_in.Pt")]),
            OrderedDict([("src", "merger.fl_out.W"), ("tgt", "duct.fl_in.W")]),
            OrderedDict([("src", "duct.fl_out.Tt"), ("tgt", "bleed.fl_in.Tt")]),
            OrderedDict([("src", "duct.fl_out.Pt"), ("tgt", "bleed.fl_in.Pt")]),
            OrderedDict([("src", "duct.fl_out.W"), ("tgt", "bleed.fl_in.W")]),
            OrderedDict([("src", "bleed.fl1_out.Tt"), ("tgt", "noz.fl_in.Tt")]),
            OrderedDict([("src", "bleed.fl1_out.Pt"), ("tgt", "noz.fl_in.Pt")]),
            OrderedDict([("src", "bleed.fl1_out.W"), ("tgt", "noz.fl_in.W")]),
            OrderedDict([("src", "fanC.fl_in.Tt"), ("tgt", "fanC.ductC.fl_in.Tt")]),
            OrderedDict([("src", "fanC.fl_in.Pt"), ("tgt", "fanC.ductC.fl_in.Pt")]),
            OrderedDict([("src", "fanC.fl_in.W"), ("tgt", "fanC.ductC.fl_in.W")]),
            OrderedDict([("src", "fanC.mech_in.XN"), ("tgt", "fanC.fan.mech_in.XN")]),
            OrderedDict([("src", "fanC.mech_in.PW"), ("tgt", "fanC.fan.mech_in.PW")]),
            OrderedDict([("src", "fanC.inwards.gh"), ("tgt", "fanC.fan.inwards.gh")]),
            OrderedDict(
                [("src", "fanC.ductC.fl_out.Tt"), ("tgt", "fanC.fan.fl_in.Tt")]
            ),
            OrderedDict(
                [("src", "fanC.ductC.fl_out.Pt"), ("tgt", "fanC.fan.fl_in.Pt")]
            ),
            OrderedDict([("src", "fanC.ductC.fl_out.W"), ("tgt", "fanC.fan.fl_in.W")]),
            OrderedDict([("src", "fanC.fan.fl_out.Tt"), ("tgt", "fanC.fl_out.Tt")]),
            OrderedDict([("src", "fanC.fan.fl_out.Pt"), ("tgt", "fanC.fl_out.Pt")]),
            OrderedDict([("src", "fanC.fan.fl_out.W"), ("tgt", "fanC.fl_out.W")]),
            OrderedDict([("src", "fanC.ductC.fl_in.Tt"), ("tgt", "fanC.ductC.merger.fl1_in.Tt")]),
            OrderedDict([("src", "fanC.ductC.fl_in.Pt"), ("tgt", "fanC.ductC.merger.fl1_in.Pt")]),
            OrderedDict([("src", "fanC.ductC.fl_in.W"), ("tgt", "fanC.ductC.merger.fl1_in.W")]),
            OrderedDict([("src", "fanC.ductC.bleed.fl2_out.Tt"), ("tgt", "fanC.ductC.merger.fl2_in.Tt")]),
            OrderedDict([("src", "fanC.ductC.bleed.fl2_out.Pt"), ("tgt", "fanC.ductC.merger.fl2_in.Pt")]),
            OrderedDict([("src", "fanC.ductC.bleed.fl2_out.W"), ("tgt", "fanC.ductC.merger.fl2_in.W")]),
            OrderedDict([("src", "fanC.ductC.bleed.fl1_out.Tt"), ("tgt", "fanC.ductC.fl_out.Tt")]),
            OrderedDict([("src", "fanC.ductC.bleed.fl1_out.Pt"), ("tgt", "fanC.ductC.fl_out.Pt")]),
            OrderedDict([("src", "fanC.ductC.bleed.fl1_out.W"), ("tgt", "fanC.ductC.fl_out.W")]),
            OrderedDict([("src", "fanC.ductC.merger.fl_out.Tt"), ("tgt", "fanC.ductC.duct.fl_in.Tt")]),
            OrderedDict([("src", "fanC.ductC.merger.fl_out.Pt"), ("tgt", "fanC.ductC.duct.fl_in.Pt")]),
            OrderedDict([("src", "fanC.ductC.merger.fl_out.W"), ("tgt", "fanC.ductC.duct.fl_in.W")]),
            OrderedDict([("src", "fanC.ductC.duct.fl_out.Tt"),
                    ("tgt", "fanC.ductC.bleed.fl_in.Tt")]),
            OrderedDict([("src", "fanC.ductC.duct.fl_out.Pt"),
                    ("tgt", "fanC.ductC.bleed.fl_in.Pt")]),
            OrderedDict([("src", "fanC.ductC.duct.fl_out.W"),
                    ("tgt", "fanC.ductC.bleed.fl_in.W")]),
                    ]

        self.expected_tree_default = OrderedDict(
            [
                ("name", "turbofan"),
                ("type", "subsystem"),
                ("subsystem_type", "group"),
                (
                    "children",
                    [
                        OrderedDict(
                            [
                                ("name", "atm"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "inlet"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "W_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        )
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "fanC"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "mech_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "XN"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "PW"),                                                                
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "ductC"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "merger"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "duct"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "outwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "bleed"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "group",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "inwards",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl_in",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "param",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl1_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "fl2_out",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "subsystem",
                                                                                ),
                                                                                (
                                                                                    "subsystem_type",
                                                                                    "component",
                                                                                ),
                                                                                (
                                                                                    "children",
                                                                                    [
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Tt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "Pt",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                        OrderedDict(
                                                                                            [
                                                                                                (
                                                                                                    "name",
                                                                                                    "W",
                                                                                                ),
                                                                                                (
                                                                                                    "type",
                                                                                                    "unknown",
                                                                                                ),
                                                                                            ]
                                                                                        ),
                                                                                    ],
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fan"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "group"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "inwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                ("children", []),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "mech_in"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "XN",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "PW",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "param",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "outwards"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                ("children", []),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "fl_out"),
                                                                ("type", "subsystem"),
                                                                (
                                                                    "subsystem_type",
                                                                    "component",
                                                                ),
                                                                (
                                                                    "children",
                                                                    [
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Tt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "Pt",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                        OrderedDict(
                                                                            [
                                                                                (
                                                                                    "name",
                                                                                    "W",
                                                                                ),
                                                                                (
                                                                                    "type",
                                                                                    "unknown",
                                                                                ),
                                                                            ]
                                                                        ),
                                                                    ],
                                                                ),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "merger"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "fl1_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "duct"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "bleed"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl1_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl2_out"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "unknown"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                        OrderedDict(
                            [
                                ("name", "noz"),
                                ("type", "subsystem"),
                                ("subsystem_type", "group"),
                                (
                                    "children",
                                    [
                                        OrderedDict(
                                            [
                                                ("name", "inwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "fl_in"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                (
                                                    "children",
                                                    [
                                                        OrderedDict(
                                                            [
                                                                ("name", "Tt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "Pt"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                        OrderedDict(
                                                            [
                                                                ("name", "W"),
                                                                ("type", "param"),
                                                            ]
                                                        ),
                                                    ],
                                                ),
                                            ]
                                        ),
                                        OrderedDict(
                                            [
                                                ("name", "outwards"),
                                                ("type", "subsystem"),
                                                ("subsystem_type", "component"),
                                                ("children", []),
                                            ]
                                        ),
                                    ],
                                ),
                            ]
                        ),
                    ],
                ),
            ]
        )

        self.expected_conns_default = [
            OrderedDict([("src", "atm.fl_out.Tt"), ("tgt", "inlet.fl_in.Tt")]),
            OrderedDict([("src", "atm.fl_out.Pt"), ("tgt", "inlet.fl_in.Pt")]),
            OrderedDict([("src", "inlet.fl_out.Tt"), ("tgt", "fanC.fl_in.Tt")]),
            OrderedDict([("src", "inlet.fl_out.Pt"), ("tgt", "fanC.fl_in.Pt")]),
            OrderedDict([("src", "inlet.fl_out.W"), ("tgt", "fanC.fl_in.W")]),
            OrderedDict([("src", "fanC.fl_out.Tt"), ("tgt", "merger.fl1_in.Tt")]),
            OrderedDict([("src", "fanC.fl_out.Pt"), ("tgt", "merger.fl1_in.Pt")]),
            OrderedDict([("src", "fanC.fl_out.W"), ("tgt", "merger.fl1_in.W")]),
            OrderedDict([("src", "bleed.fl2_out.Tt"), ("tgt", "merger.fl2_in.Tt")]),
            OrderedDict([("src", "bleed.fl2_out.Pt"), ("tgt", "merger.fl2_in.Pt")]),
            OrderedDict([("src", "bleed.fl2_out.W"), ("tgt", "merger.fl2_in.W")]),
            OrderedDict([("src", "merger.fl_out.Tt"), ("tgt", "duct.fl_in.Tt")]),
            OrderedDict([("src", "merger.fl_out.Pt"), ("tgt", "duct.fl_in.Pt")]),
            OrderedDict([("src", "merger.fl_out.W"), ("tgt", "duct.fl_in.W")]),
            OrderedDict([("src", "duct.fl_out.Tt"), ("tgt", "bleed.fl_in.Tt")]),
            OrderedDict([("src", "duct.fl_out.Pt"), ("tgt", "bleed.fl_in.Pt")]),
            OrderedDict([("src", "duct.fl_out.W"), ("tgt", "bleed.fl_in.W")]),
            OrderedDict([("src", "bleed.fl1_out.Tt"), ("tgt", "noz.fl_in.Tt")]),
            OrderedDict([("src", "bleed.fl1_out.Pt"), ("tgt", "noz.fl_in.Pt")]),
            OrderedDict([("src", "bleed.fl1_out.W"), ("tgt", "noz.fl_in.W")]),
            OrderedDict([("src", "fanC.fl_in.Tt"), ("tgt", "fanC.ductC.fl_in.Tt")]),
            OrderedDict([("src", "fanC.fl_in.Pt"), ("tgt", "fanC.ductC.fl_in.Pt")]),
            OrderedDict([("src", "fanC.fl_in.W"), ("tgt", "fanC.ductC.fl_in.W")]),
            OrderedDict([("src", "fanC.mech_in.XN"), ("tgt", "fanC.fan.mech_in.XN")]),
            OrderedDict([("src", "fanC.mech_in.PW"), ("tgt", "fanC.fan.mech_in.PW")]),
            OrderedDict(
                [("src", "fanC.ductC.fl_out.Tt"), ("tgt", "fanC.fan.fl_in.Tt")]
            ),
            OrderedDict(
                [("src", "fanC.ductC.fl_out.Pt"), ("tgt", "fanC.fan.fl_in.Pt")]
            ),
            OrderedDict([("src", "fanC.ductC.fl_out.W"), ("tgt", "fanC.fan.fl_in.W")]),
            OrderedDict([("src", "fanC.fan.fl_out.Tt"), ("tgt", "fanC.fl_out.Tt")]),
            OrderedDict([("src", "fanC.fan.fl_out.Pt"), ("tgt", "fanC.fl_out.Pt")]),
            OrderedDict([("src", "fanC.fan.fl_out.W"), ("tgt", "fanC.fl_out.W")]),
            OrderedDict(
                [("src", "fanC.ductC.fl_in.Tt"), ("tgt", "fanC.ductC.merger.fl1_in.Tt")]
            ),
            OrderedDict(
                [("src", "fanC.ductC.fl_in.Pt"), ("tgt", "fanC.ductC.merger.fl1_in.Pt")]
            ),
            OrderedDict(
                [("src", "fanC.ductC.fl_in.W"), ("tgt", "fanC.ductC.merger.fl1_in.W")]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.bleed.fl2_out.Tt"),
                    ("tgt", "fanC.ductC.merger.fl2_in.Tt"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.bleed.fl2_out.Pt"),
                    ("tgt", "fanC.ductC.merger.fl2_in.Pt"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.bleed.fl2_out.W"),
                    ("tgt", "fanC.ductC.merger.fl2_in.W"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.bleed.fl1_out.Tt"),
                    ("tgt", "fanC.ductC.fl_out.Tt"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.bleed.fl1_out.Pt"),
                    ("tgt", "fanC.ductC.fl_out.Pt"),
                ]
            ),
            OrderedDict(
                [("src", "fanC.ductC.bleed.fl1_out.W"), ("tgt", "fanC.ductC.fl_out.W")]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.merger.fl_out.Tt"),
                    ("tgt", "fanC.ductC.duct.fl_in.Tt"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.merger.fl_out.Pt"),
                    ("tgt", "fanC.ductC.duct.fl_in.Pt"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.merger.fl_out.W"),
                    ("tgt", "fanC.ductC.duct.fl_in.W"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.duct.fl_out.Tt"),
                    ("tgt", "fanC.ductC.bleed.fl_in.Tt"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.duct.fl_out.Pt"),
                    ("tgt", "fanC.ductC.bleed.fl_in.Pt"),
                ]
            ),
            OrderedDict(
                [
                    ("src", "fanC.ductC.duct.fl_out.W"),
                    ("tgt", "fanC.ductC.bleed.fl_in.W"),
                ]
            ),
        ]

    def test_model_viewer_has_correct_data_from_problem(self):
        """
        Verify that the correct model structure data exists when stored as compared
        to the expected structure, using the SellarStateConnection model.
        """
        p = ComplexTurbofan("turbofan")

        model_viewer_data = _get_viewer_data(p, include_wards=True)
        tree_json = model_viewer_data["tree"]
        conns_json = model_viewer_data["connections_list"]

        assert self.expected_tree == tree_json
        assert self.expected_conns == conns_json

    def test_model_viewer_takes_into_account_include_wards(self):
        """
        Test that model viewer does not show include_wards by default.
        """
        p = ComplexTurbofan("turbofan")

        model_viewer_data = _get_viewer_data(p, include_wards=False)
        tree_json = model_viewer_data["tree"]
        conns_json = model_viewer_data["connections_list"]

        assert self.expected_tree_default == tree_json
        assert self.expected_conns_default == conns_json

    def test_view_model_from_problem(self):
        """
        Test that an n2 html file is generated from a Problem.
        """
        p = ComplexTurbofan("turbofan")

        view_model(p, outfile=self.problem_filename, show_browser=False)

        # Check that the html file has been created and has something in it.
        self.assertTrue(
            os.path.isfile(self.problem_html_filename),
            (self.problem_html_filename + " is not a valid file."),
        )
        self.assertGreater(os.path.getsize(self.problem_html_filename), 100)




import pytest

import math
import numpy as np
import warnings

from cosapp.core.eval_str import EvalString
from cosapp.systems import System
from cosapp.utils.testing import assert_keys


def test_EvalString__init__(eval_context):
    e = EvalString('3.14159', eval_context)
    assert e.eval_context is eval_context
    assert e.constant
    assert len(e.locals) == 0

    e = EvalString('22 + a', eval_context)
    assert e.eval_context is eval_context
    assert not e.constant
    assert_keys(e.locals, 'a')


@pytest.mark.parametrize("f", [
        func_name for func_name in EvalString.available_symbols()
        if not func_name.startswith('u')
])
def test_EvalString_exec_comp_value(f, ufunc_test_data):
    class Test(System):
        def setup(self, **kwargs):
            for arg_name, arg_value in kwargs.items():
                self.add_inward(arg_name, **arg_value)

    test_data = ufunc_test_data[f]

    if 'check_func' in test_data:
        check_args = []
        try:
            check_args.append(test_data['args']['x']['value'])
        except:
            pass
        try:
            check_args.append(test_data['args']['y']['value'])
        except:
            pass
        expected = test_data['check_func'](*check_args)
    else:
        expected = test_data['check_val']

    with warnings.catch_warnings():
        warnings.simplefilter('error')  # 'default', 'ignore', or 'error'
        s = Test('tmp', **test_data.get('args', {}))
        eval_str = EvalString(test_data['str'], s)

    assert eval_str.eval() == pytest.approx(expected, rel=1e-14)


@pytest.mark.skip(reason='TODO')
def test_EvalString_residue_as_context(self):
    pass


@pytest.mark.parametrize("expression, expected", [
    (1, 1),
    (-1, -1),
    (0.2, 0.2),
    ("--1", 1),
    ("0.2", 0.2),
    ("0.2 / 0.5", 0.4),
    ("cos(pi)", -1),
    ("sin(pi / 2)", 1),
    ("log(exp(2))", 2),
    ("log(e)", 1),
    (None, None),
    ("None", None),
    (dict(), dict()),
    ("[]", list()),
    (True, True),
    (False, False),
    ("True", True),
    ("False", False),
    ("0.3 < 0.7", True),
    ("0.3 <= 0.3", True),
    ("0.3 < 0.7 < 0.6", False),
    ("0.3 + 0.7 == 1", True),
    ("[0., 0., 0.]", [0., 0., 0.]),
    ("[0.] * 3", [0., 0., 0.]),
    ("array([0, 0, 0], dtype=float)", [0., 0., 0.]),
    ("ones(4, dtype=int)", [1, 1, 1, 1]),
    ("ones(2, dtype=bool)", [True, True]),
    (repr(np.zeros(3)), [0., 0., 0.]),
    ("zeros(3)", [0., 0., 0.]),
    ("ones(3)", [1., 1., 1.]),
    (np.zeros(3), [0., 0., 0.]),
    (np.zeros((2, 5, 3)), np.zeros((2, 5, 3))),
    (np.ones(3), [1., 1., 1.]),
    ("-e", -math.e),
    ("9 + 3 + 6", 18),
    ("9 + 3 / 11", 9 + 3 / 11),
    ("(9 + 3) / 11", 12 / 11),
    ("(9 + 3)", 12),
    ("9 - 12 - 6", -9),
    ("9 - (12 - 6)", 9 - (12 - 6)),
    ("2 * 3.14159",  2 * 3.14159),
    ("3.1415926535 * 3.1415926535 / 10", 3.1415926535 * 3.1415926535 / 10),
    ("pi * pi / 10", math.pi * math.pi / 10),
    ("pi**2 / 10", math.pi**2 / 10),
    ("pi**2", math.pi ** 2),
    ("round(pi**2)", round(math.pi ** 2)),
    ("6.02E23 * 8.048", 6.02e23 * 8.048),
    ("e / 3", math.e / 3),
    ("round(e)", round(math.e)),
    ("round(-e)", round(-math.e)),
    ("e**pi", math.e ** math.pi),
    ("2**3**2", 2 ** 3 ** 2),
    ("2**3+2", 2 ** 3 + 2),
    ("2**9", 2 ** 9),
    ("{1, 2, 3, 3, 2, }", {1, 2, 3}),
])
def test_EvalString_constant_expr(eval_context, expression, expected):
    """Test expressions expected to be interpreted as constant"""
    s = EvalString(expression, eval_context)
    assert s.eval_context is eval_context
    assert s.constant
    if expected is None:
        assert s.eval() is None
    else:
        assert s.eval() == pytest.approx(expected, rel=1e-14)


@pytest.mark.parametrize("expression, expected", [
    ("-2 * g", dict(value=pytest.approx(-2 * 9.80665, rel=1e-14), constant=True)),
    ("1e-23 * NA + g", dict(value=pytest.approx(15.828790), constant=True)),
    ("g * x", dict(constant=False)),
])
def test_EvalString_with_constants(eval_context, expression, expected):
    """Test expressions with system constants, expected to be interpreted as constant"""
    s = EvalString(expression, eval_context)
    assert s.eval_context is eval_context
    assert s.constant == expected.get('constant', False)
    if 'value' in expected:
        assert s.eval() == expected['value']


@pytest.mark.parametrize("expression, expected", [
    ("norm(x, inf)", 3.14),
    ("0.2 * a / 0.5", 0.8),
    ("log10(x[0])", -1),
    ("a + b - x[1]", 2.7),
    ("a - a + 1", 1),
    ("9 + sub.in_.q / 10", 9.5),
    ("9 + sin(sub.in_.q / 11)", 9 + math.sin(5 / 11)),
    ("out.q - sub.in_.q", -4.5),
    ("concatenate((x, [-a, sub.in_.q]))", [0.1, -0.2, -3.14, -2, 5]),
    ])
def test_EvalString_nonconstant_expr(eval_context, expression, expected):
    """Test expressions expected to be interpreted as non-constant"""
    s = EvalString(expression, eval_context)
    assert s.eval_context is eval_context
    assert not s.constant
    assert s.eval() == pytest.approx(expected)


@pytest.mark.parametrize("expression, exception", [
    ("a = 2.0", SyntaxError),
    ("cos(pi", SyntaxError),
    ("1 + _", NameError),
    ("1 / 0", ZeroDivisionError),
    ("foo.bar(1)", NameError),
    ("a + ['v', True]", TypeError),
    ("out.monkey", AttributeError),
    ("sub.out.q", AttributeError),
    ("", ValueError),
    ("  ", ValueError),
    ])
def test_EvalString_erroneous_expr(eval_context, expression, exception):
    """Test erroneous expressions expected to raise an exception at instantiation"""
    with pytest.raises(exception):
        EvalString(expression, eval_context)


@pytest.mark.parametrize("expression, expected", [
    ("1", "1"),
    ("1.2", "1.2"),
    (8, "8"),
    (1.3, "1.3"),
    ([0.1, 0.2, 0.3], "[0.1, 0.2, 0.3]"),
    (np.r_[0.1, 0.2, 0.3], "array([0.1, 0.2, 0.3])"),
    ])
def test_EvalString_str_repr(eval_context, expression, expected):
    s = EvalString(expression, eval_context)
    assert str(s) == expected
    assert repr(s) == expected.join("''")  # Note: same behaviour for built-in Python strings


@pytest.mark.parametrize("expression, patterns, results", [
    ("norm(x, inf)", ["norm", "inf", "x", "foo"], [True] * 3 + [False]),
    ("0.2 * a / 0.5", ["0.", "a", "x"], [True, True, False]),
    ("x", ["x", "Y", "foo"], [True, False, False]),
    ("[a, 2]", ["a", "2", "1", "foo"], [True, True, False, False]),
    ("[1, 2]", ["a", "2", "1", "1, 2"], [False, True, True, True]),
    ])
def test_EvalString_contains(eval_context, expression, patterns, results):
    """Test expressions expected to be interpreted as non-constant"""
    s = EvalString(expression, eval_context)
    for pattern, expected in zip(patterns, results):
        assert (pattern in s) == expected


@pytest.mark.parametrize("var", [
    "a", "b", "out.q", "sub.z", "sub.in_.q", ])
def test_EvalString_multiple_eval(eval_context, var):
    s = EvalString(var, eval_context)
    assert s.eval() == eval_context[var]
    for eval_context[var] in [0, 0.1, -22]:
        assert s.eval() == eval_context[var]


@pytest.mark.parametrize("expression, expected", [
    ("norm(x, inf)", "norm(x, inf)"),
    ("123.456", "123.456"),
    (123.456, "123.456"),
    ([1, 2, 3, ], "[1, 2, 3]"),
    ((1, 2, 3, ), "(1, 2, 3)"),
    ({1, 2, 3}, "{1, 2, 3}"),
    ("{1, 2, 3, 3, 2, }", "{1, 2, 3, 3, 2, }"),
    (np.r_[0.1, 0.2, 0.3], "array([0.1, 0.2, 0.3])"),
    (np.array([0, 0, 0], dtype=float), "array([0., 0., 0.])"),
    (np.ones(2, dtype=int), "array([1, 1])"),
    (None, "None"),
    ])
def test_EvalString_string(eval_context, expression, expected):
    assert EvalString.string(expression) == expected
    s = EvalString(expression, eval_context)
    assert EvalString.string(s) == expected



"""Utility functions for testing purposes"""
import numpy
from contextlib import contextmanager


def assert_keys(dictionary, *keys):
    """Utility function to test dictionary keys"""
    if set(dictionary.keys()) != set(keys):
        actual = list(dictionary.keys())
        raise KeyError(f"actual: {actual}; expected: {list(keys)}") 


def assert_all_type(collection, dtype):
    """Asserts that all the elements of a collecion are of a given type"""
    def assert_type(name, value):
        if not isinstance(value, dtype):
            actual = type(value).__name__
            raise TypeError(
                f"{name} is of type {actual}; expected {dtype.__name__}"
            ) 
    if isinstance(collection, dict):
        for key, value in collection.items():
            assert_type(f"element with key '{key}'", value)
    else:
        for i, value in enumerate(collection):
            assert_type(f"element #{i}", value)

def rel_error(actual, expected):
    res = lambda a, b: abs(a) if b == 0 else abs(a / b - 1)
    if isinstance(expected, (list, tuple, numpy.ndarray)):
        return numpy.array([res(a, b) for a, b in zip(actual, expected)])
    else:
        return res(actual, expected)


def get_args(*args, **kwargs):
    """Utility function to collect args and kwargs in a tuple"""
    return args, kwargs


@contextmanager
def no_exception():
    """Context manager to assert that a block does not raise any exception"""
    try:
        yield

    except Exception as error:
        raise AssertionError(f"Unexpected exception raised: {error!r}")


@contextmanager
def not_raised(ExpectedException):
    """Context manager to assert that a block does not raise `ExpectedException`"""
    # https://gist.github.com/oisinmulvihill/45c14271fad7794a4a52516ecb784e69
    try:
        yield

    except ExpectedException as error:
        raise AssertionError(f"Raised {error!r} exception when it should not!")

    except Exception as error:
        raise AssertionError(f"Unexpected exception raised: {error!r}")



"""Recorder in `pandas.DataFrame`."""
import copy
import numpy
import pandas
from typing import Any, List, Optional, Union

from cosapp.recorders.recorder import BaseRecorder


class DataFrameRecorder(BaseRecorder):
    """Record data into a pandas.DataFrame.

    Matching pattern are case sensitive and support the following special patterns:

    ========  ================================
    Pattern   Meaning
    ========  ================================
    `*`       matches everything
    `?`       matches any single character
    `[seq]`   matches any character in seq
    `[!seq]`  matches any character not in seq
    ========  ================================

    Excluding pattern are shadowing includes one; e.g. if `includes='*port_in.*' and `excludes='*.Pt'`, for a port
    having variables named `Tt` and `Pt`, only `Tt` will be recorded.

    Parameters
    ----------
    includes : str or list of str, optional
        Variables matching these patterns will be included; default `'*'` (i.e. all variables).
    excludes : str or list of str or None, optional
        Variables matching these patterns will be excluded; default `None` (i.e. nothing is excluded).
    numerical_only : bool, optional
        Keep only numerical variables (i.e. number or numerical vector); default False.
    section : str, optional
        Current section name; default `''`.
    precision : int, optional
        Precision digits when writing floating point number; default 9 (i.e. 10 figures will be written).
    hold : bool, optional
        Append the new data or not; default `False`.
    raw_output : bool, optional
        Raw output; default `False`.

    ..note::
        Do not mention `inwards` or `outwards` in `includes` or `excludes` list. Otherwise you may not record the wanted
        variables.
    """

    def __init__(
        self,
        includes: Union[str, List[str]] = "*",
        excludes: Optional[Union[str, List[str]]] = None,
        # metadata: Optional[Union[str, List[str]]] = None,
        numerical_only = False,
        section = "",
        precision = 9,
        hold = False,
        raw_output = True,
    ):
        super().__init__(
            includes, excludes, numerical_only, section, precision, hold, raw_output
        )
        # Temporary storage
        self.__buffer = list()  # type: List[List[Any]]

    def export_data(self) -> pandas.DataFrame:
        """Export recorded results into a pandas.DataFrame object."""
        # According to DataFrame documentation, it is more efficient to store in a list than create the DataFrame
        headers = [
            self.SPECIALS.section,
            self.SPECIALS.status,
            self.SPECIALS.code,
            self.SPECIALS.reference,
        ]
        varlist = self.field_names()
        if self._raw_output:
            headers.extend(varlist)
        else:
            # Add units in column headers
            headers.extend(
                map(
                    lambda v: f"{v[0]} [{v[1]}]",
                    zip(varlist, self._get_units(varlist)),
                )
            )
        return pandas.DataFrame(self.__buffer, columns=headers)

    @property
    def _raw_data(self) -> List[List[Any]]:
        """Return a raw/unformatted version of records

        Returns
        -------
        List[List[Any]]
            Records of `watched_object` for variables given by method `field_names()`
        """
        return self.__buffer

    def start(self):
        """Initialize recording support."""
        super().start()
        if not self.hold:
            self.__buffer.clear()

    def formatted_data(self) -> List[Any]:
        """Collect recorded data from watched object into a list."""
        line = []
        names = self.field_names()
        values = self.collected_data()
        for name, value in zip(names, values):
            try:
                line.append(copy.deepcopy(value))
            except copy.Error:
                context = self.watched_object
                if name in context:
                    varname = f"{context.name}.{name}"
                else:
                    varname = f"{context.name}[{name}]"
                raise TypeError(
                    f"Cannot record {varname}: DataFrameRecorder objects can only capture deep-copyable variables"
                )
        return line

    def _record(self, line: List[Any]) -> None:
        self.__buffer.append(line)

    def exit(self):
        """Close recording session."""
        pass

    def clear(self):
        """Clear all previously stored data."""
        self.__buffer.clear()
        super().clear()



"""Customized log handler for CoSApp simulation."""
import io
import logging
import sys
from contextlib import contextmanager
from enum import Enum, IntEnum
from logging.handlers import RotatingFileHandler
from numbers import Number
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Iterable,
    List,
    Mapping,
    Optional,
    Type,
    Union,
)

from cosapp.utils.helpers import check_arg


root_logger = logging.getLogger()
logger = logging.getLogger(__name__)


DEFAULT_STREAM = object()


class LogFormat(Enum):
    """Expected format of the log message.
    
    RAW : Raw text
    """

    RAW = 0


class LogLevel(IntEnum):
    """CoSApp log level.
    
    FULL_DEBUG : Detailed debug log
    DEBUG : Debug log
    INFO : Information log
    WARNING : Warning log
    ERROR : Error log
    CRITICAL : Critical log
    """

    FULL_DEBUG = logging.DEBUG - 1
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


VERBOSE_LEVEL = LogLevel.DEBUG


class HandlerWithContextFilters:
    """Add method to handlers for CoSApp log message."""

    def __init__(self):
        self.contextual_filters: List[FilterWithContext] = list()

    def _set_contextual_filters(self, filters: Iterable[logging.Filter]) -> None:
        self.contextual_filters = list(
            filter(lambda f: isinstance(f, FilterWithContext), filters)
        )

    def log(
        self,
        level: LogLevel,
        msg: str,
        name: str = "",
        fn: str = "",
        lino: int = 0,
        args: tuple = (),
        exc_info=None,
        func: Optional[str] = None,
        extra: Optional[Mapping[str, Any]] = None,
        sinfo: Optional[str] = None,
    ) -> None:
        """Helper function to publish log message with this handler.

        Parameters
        ----------
        level : LogLevel
            The numeric level of the logging event (one of DEBUG, INFO etc.) Note that
            this is converted to two attributes of the LogRecord: levelno for the numeric
            value and levelname for the corresponding level name.
        msg : str
            The event description message, possibly a format string with placeholders
            for variable data.
        name : str, optional
            The name of the logger used to log the event represented by this LogRecord.
            Note that this name will always have this value, even though it may be
            emitted by a handler attached to a different (ancestor) logger.
        pathname : str, optional
            The full pathname of the source file where the logging call was made.
        lineno : int, optional
            The line number in the source file where the logging call was made.
        args : tuple
            Variable data to merge into the msg argument to obtain the event description.
        exc_info : optional
            An exception tuple with the current exception information, or None if no
            exception information is available.
        func : str, optional
            The name of the function or method from which the logging call was invoked.
        sinfo: str, optional
            A text string representing stack information from the base of the stack in
            the current thread, up to the logging call.
        """
        if isinstance(self, logging.Handler):
            if self.level <= level:
                record = logger.makeRecord(
                    name or logger.name,
                    level,
                    fn,
                    0,
                    msg,
                    args,
                    exc_info,
                    func,
                    extra,
                    sinfo,
                )
                self.handle(record)
        else:
            raise NotImplementedError(f"{self} cannot handle log messages.")

    def needs_handling(self, record: logging.LogRecord) -> bool:
        """Test the record to see if it needs to be processed or not.
        
        Parameters
        ----------
        record : logging.LogRecord
            Log record

        Returns
        -------
        bool
            Is the record to be processed?
        """
        context = getattr(record, "context", None)
        result = True
        if context is not None:
            activation = getattr(record, "activate", None)

            if activation == True:
                for f in self.contextual_filters:
                    f.current_context = context

            result = False
            if isinstance(context, LoggerContext):
                result = context.log_debug_message(self, record)

            if activation == False:
                for f in self.contextual_filters:
                    f.current_context = context

        return result


class LoggerContext:
    """Interface for context object to connect to the logging system."""

    __slots__ = ()

    CONTEXT_ENTER_MESSAGE: ClassVar[str] = "Entering "
    CONTEXT_EXIT_MESSAGE: ClassVar[str] = "Exiting "

    @contextmanager
    def log_context(self, suffix: str = "") -> None:
        """Set this object as the context for the logger.

        Parameters
        ----------
        suffix : str, optional
            Suffix text to append to the log message
        """
        try:
            msg = LoggerContext.CONTEXT_ENTER_MESSAGE + repr(self) + suffix
            logger.log(VERBOSE_LEVEL, msg, extra={"activate": True, "context": self})
            yield
        finally:
            msg = LoggerContext.CONTEXT_EXIT_MESSAGE + repr(self) + suffix
            logger.log(VERBOSE_LEVEL, msg, extra={"activate": False, "context": self})

    def log_debug_message(
        self,
        handler: HandlerWithContextFilters,
        record: logging.LogRecord,
        format: LogFormat = LogFormat.RAW,
    ) -> bool:
        """Callback method on the context object to log more detailed information.

        This method will be called by the log handler when :py:meth:`~cosapp.utils.logging.LoggerContext.log_context`
        is active if the logging level is lower or equals to VERBOSE_LEVEL. It allows
        the object to send additional log message to help debugging a simulation.

        Parameters
        ----------
        handler : HandlerWithContextFilters
            Log handler on which additional message should be published.
        record : logging.LogRecord
            Log record
        format : LogFormat
            Format of the message

        Returns
        -------
        bool
            Should the provided records be logged?
        """
        return True


class FilterWithContext(logging.Filter):
    """Interface to add a context on an object."""

    def __init__(self):
        super().__init__()
        self.__context = None

    @property
    def current_context(self) -> LoggerContext:
        """LoggerContext : Current context"""
        return self.__context

    @current_context.setter
    def current_context(self, context: LoggerContext) -> None:
        self.__context = context
        self._set_context()

    def _set_context(self) -> None:
        """Hook method called by current_context setter."""
        pass


class TimeFilter(FilterWithContext):
    """Log record filter depending on the current time of a context.
    
    Parameters
    ----------
    start_time : Number
        Time from which debug log will be recorded; default all time steps
    """

    def __init__(self, start_time: Number):
        super().__init__()
        self.__start_time = start_time
        self.__current_time = start_time

    def _set_context(self) -> None:
        """Update current time with the one of the context."""
        self.__current_time = getattr(self.current_context, "time", self.__current_time)

    def filter(self, record: logging.LogRecord) -> int:
        """Is the specified record to be logged? Returns zero for no, nonzero for yes.
        If deemed appropriate, the record may be modified in-place by this method.
        
        Parameters
        ----------
        record : logging.LogRecord
            Log record to test
        
        Returns
        -------
        int
            Non-zero if the record is to be logged.
        """
        return int(
            record.levelno > VERBOSE_LEVEL or self.__current_time >= self.__start_time
        )


class ContextFilter(FilterWithContext):
    def __init__(self, context: Optional[str] = None):
        super().__init__()

        self.__filter_context: Optional[str] = context  # Filter context
        self.__active: bool = False
        self.__context_type: Optional[Type] = None

        # Set context filter logic
        if self.__filter_context is None:
            self._context_filter = lambda record: 1
        else:
            self._context_filter = lambda record: int(
                record.levelno > VERBOSE_LEVEL or self.__active
            )

    def filter(self, record: logging.LogRecord) -> int:
        """Is the specified record to be logged? Returns zero for no, nonzero for yes.
        If deemed appropriate, the record may be modified in-place by this method.
        
        Parameters
        ----------
        record : logging.LogRecord
            Log record to test
        
        Returns
        -------
        int
            Non-zero if the record is to be logged.
        """
        return self._context_filter(record)

    def _set_context(self) -> None:
        context = self.current_context
        context_name = getattr(context, "name", None)

        if self.__context_type is not None:
            if context_name == self.__filter_context:
                self.__context_type = None
                self.__active = False
            else:
                self.__active = isinstance(context, self.__context_type)

        elif context_name == self.__filter_context:
            from cosapp.drivers import Driver
            from cosapp.systems import System

            if isinstance(context, System):
                self.__context_type = System
            elif isinstance(context, Driver):
                self.__context_type = Driver
            else:
                self.__context_type = type(context)
            self.__active = True


class FileLogHandler(RotatingFileHandler, HandlerWithContextFilters):
    """Special RotatingFileHandler for CoSApp log message.

    Parameters
    ----------
    filename : str or Path, optional
        Log filename; default "cosapp_trace.log"
    backupCount : int, optional
        Number of backup log files; default 5
    encoding : str, optional
        File encoding to be enforced
    """

    def __init__(
        self,
        filename: Union[str, Path] = "cosapp_trace.log",
        backupCount: int = 5,
        encoding: Optional[str] = None,
    ) -> None:
        RotatingFileHandler.__init__(
            self, filename, backupCount=backupCount, encoding=encoding, delay=True
        )
        HandlerWithContextFilters.__init__(self)

    def addFilter(self, filter):
        """Adds the specified filter filter to this handler."""
        super().addFilter(filter)
        self._set_contextual_filters(self.filters)

    def removeFilter(self, filter):
        """Removes the specified filter filter from this handler."""
        super().removeFilter(filter)
        self._set_contextual_filters(self.filters)

    def handle(self, record: logging.LogRecord) -> bool:
        """Conditionally emits the specified logging record, depending on filters which
        may have been added to the handler. Wraps the actual emission of the record
        with acquisition/release of the I/O thread lock.
        
        Parameters
        ----------
        record : logging.LogRecord
            Log record

        Returns
        -------
        bool
            Is the record processed?
        """
        return self.needs_handling(record) and super().handle(record)


class StreamLogHandler(logging.StreamHandler, HandlerWithContextFilters):
    """Special StreamHandler for CoSApp log message."""

    def __init__(self, stream: io.TextIOBase = DEFAULT_STREAM) -> None:
        if stream is DEFAULT_STREAM:
            stream = sys.stdout
        logging.StreamHandler.__init__(self, stream=stream)
        HandlerWithContextFilters.__init__(self)

    def addFilter(self, filter):
        """Adds the specified filter filter to this handler."""
        super().addFilter(filter)
        self._set_contextual_filters(self.filters)

    def removeFilter(self, filter):
        """Removes the specified filter filter from this handler."""
        super().removeFilter(filter)
        self._set_contextual_filters(self.filters)

    def handle(self, record: logging.LogRecord) -> bool:
        """Conditionally emits the specified logging record, depending on filters which
        may have been added to the handler. Wraps the actual emission of the record
        with acquisition/release of the I/O thread lock.
        
        Parameters
        ----------
        record : logging.LogRecord
            Log record

        Returns
        -------
        bool
            Is the record processed?
        """
        return self.needs_handling(record) and super().handle(record)


def rollover_logfile() -> None:
    """Rollover logfile of CoSApp LogHandler."""
    root_logger = logging.getLogger()

    for handler in root_logger.handlers:
        if isinstance(handler, RotatingFileHandler):
            handler.doRollover()


def set_log(
    filename: Union[str, Path, None] = "cosapp_trace.log",
    stream: Optional[io.TextIOBase] = DEFAULT_STREAM,
    level: int = LogLevel.INFO,
    context: Optional[str] = None,
    start_time: Optional[Number] = None,
    format: str = "%(message)s",
    encoding: Optional[str] = None,
    backupCount: int = 5,
) -> None:
    """Set the CoSApp simulation log behavior.

    If `backupCount` is nonzero, at most `backupCount` files will be kept, and if more
    would be created when rollover occurs, the oldest one is deleted.

    The system will save old log files by appending extensions to the filename. The
    extensions are date-and-time based, using the strftime format `%Y-%m-%d_%H-%M-%S`.

    By default the log messages are written to a file (specified by its `filename`) and
    to a stream. Set either `filename` or `stream` to deactivate the corresponding log
    handler.

    Parameters
    ----------
    filename : str or Path or None, optional
        Log filename; default "cosapp_trace.log"
    stream : io.TextIOBase or None, optional
        Log stream; default ``sys.stdout``
    level : int or LogLevel, optional
        Log level; default LogLevel.INFO
    context : str or None, optional
        Context on which to focus the log message; default None
    start_time : Number or None, optional
        Time from which debug log will be recorded; default all time steps
    format : str, optional
        Log record format; default "%(message)s" - for the available attributes (see https://docs.python.org/3/library/logging.html#logrecord-attributes)
    encoding : str, optional
        File encoding to be enforced
    backupCount : int, optional
        Number of backup log files; default 5
    """
    nonetype = type(None)
    check_arg(filename, "filename", (str, Path, nonetype))
    if stream is not DEFAULT_STREAM:
        check_arg(stream, "stream", (io.TextIOBase, nonetype))
    check_arg(level, "level", (int, LogLevel))
    check_arg(context, "context", (str, nonetype))
    check_arg(start_time, "start_time", (Number, nonetype))
    check_arg(format, "format", str)
    check_arg(encoding, "encoding", (str, nonetype))
    check_arg(backupCount, "backupCount", int, lambda v: v >= 0)

    root_logger.setLevel(level)

    for handler in root_logger.handlers[:]:
        if isinstance(handler, HandlerWithContextFilters):
            handler.close()  # Be sure to close the file descriptor
            root_logger.removeHandler(handler)

    def add_handler(h):
        fmt = logging.Formatter(format)
        h.setFormatter(fmt)
        h.setLevel(level)
        if context is not None:
            h.addFilter(ContextFilter(context))
        if start_time is not None:
            h.addFilter(TimeFilter(start_time))
        root_logger.addHandler(h)

    handlers = list()
    if filename is not None:
        handlers.append(
            FileLogHandler(filename, backupCount=backupCount, encoding=encoding)
        )

    if stream is not None:
        handlers.append(StreamLogHandler(stream))

    for handler in handlers:
        add_handler(handler)

    if len(handlers) == 0:
        logger.warning("No CoSApp log handlers added.")



from typing import Optional, Tuple
from collections.abc import Collection

import numpy

from .helpers import check_arg


# TODO should we reintegrate it as Boundary method as it is now the only used place
def get_indices(
    syst: "cosapp.core.module.Module",
    name: str,
) -> Tuple[str, Optional[numpy.ndarray]]:
    """Decompose a variable specification into its name and its mask.

    Parameters
    ----------
    syst : Module
        Module to which belong the variable
    name : str
        Variable specification

    Returns
    -------
    Tuple[str, Optional[numpy.ndarray[bool]]]
        Tuple (variable name, variable mask or None)
    """
    check_arg(name, 'name', str)
    if "[" in name:  # User want to use a part of a sequence
        var_name, selector = name.split("[", maxsplit=1)
        selector = "[" + selector
    else:
        var_name = name
        selector = None

    if var_name not in syst:
        raise AttributeError(
            f"Variable {var_name!r} is not known in system {syst.name!r}."
        )

    value = syst[var_name]
    if selector:
        # Check value is an array
        if not (
            isinstance(value, numpy.ndarray)
            and numpy.issubdtype(value.dtype, numpy.number)
            and value.size > 1
        ):
            raise TypeError(
                f"Only non-empty numpy array can be partially selected; got '{syst.name}.{name}'."            )
        mask = numpy.zeros(value.shape, dtype=bool)
        # Set the mask using the selector
        try:
            exec(f"mask{selector} = True", {}, {"mask": mask})
        except (SyntaxError, IndexError) as err:
            raise type(err)(
                "Selection {!r} for variable '{}.{}' is not valid: {!s}".format(
                    selector, syst.name, var_name, err
                )
            )
    else:
        mask = None
        if isinstance(value, Collection) and not isinstance(value, str):
            mask = numpy.ones_like(value, dtype=bool)

    return var_name, mask



import json

import numpy
import pytest

from ..distribution import Distribution
from ..triangular import Triangular


def test_Triangular___init__():
    distrib = Triangular(-0.5, 1, 2)

    assert isinstance(distrib, Distribution)
    assert isinstance(distrib, Triangular)

    assert distrib.worst == -0.5
    assert distrib.likely == 1
    assert distrib.best == 2

    with pytest.raises(ValueError):
        Triangular(worst=9.0, likely=6.0, best=7.0)

    with pytest.raises(ValueError):
        Triangular(worst=1.0, likely=10.0, best=7.0)

    distrib2 = Triangular(worst=9.0, likely=6.0, best=5.0, pworst=0.0, pbest=0.0)
    for a, v in (
        {"worst": 9.0, "likely": 6.0, "best": 5.0, "pworst": 0.0, "pbest": 0.0}
    ).items():
        assert getattr(distrib2, a) == v

    with pytest.raises(
        ValueError,
        match=r"Likely value not within distribution bounds: \d+\.\d+ <= \d+\.\d+ <= \d+\.\d+\.",
    ):
        Triangular(4.0, 2.0, 6.0)


def test_Triangular___json__():
    distrib = Triangular(worst=9.0, likely=6.0, best=5.0, pworst=0.2, pbest=0.1)

    assert distrib.__json__() == {
        "worst": 9.0,
        "pworst": 0.2,
        "best": 5.0,
        "pbest": 0.1,
        "likely": 6.0,
    }


def test_Triangular_likely():
    distrib = Triangular(8.0, 9.0, 11.0)
    distrib.likely = 10.0
    assert 10.0 == distrib.likely

    distrib = Triangular(12.0, 10.0, 9.0, pworst=0.0, pbest=0.0)

    with pytest.raises(ValueError):
        distrib.likely = 8.0

    distrib.likely = 9.1
    assert 9.1 == distrib.likely


def test_Triangular_draw():
    # Symmetric case
    distrib = Triangular(8.0, 9.0, 10.0)

    assert distrib.draw(distrib.pworst) == pytest.approx(distrib.worst, abs=1e-7)
    assert distrib.draw(1.0 - distrib.pbest) == pytest.approx(distrib.best, abs=1e-7)
    assert distrib.draw(0.5) == pytest.approx(distrib.likely, abs=1e-7)

    # Unsymmetric case
    distrib = Triangular(8.0, 9.0, 10.0, pworst=0.2, pbest=0.1)

    assert distrib.draw(distrib.pworst) == pytest.approx(distrib.worst, abs=1e-7)
    assert distrib.draw(1.0 - distrib.pbest) == pytest.approx(distrib.best, abs=1e-7)

    loc = distrib.draw(0.0)
    scale = distrib.draw(1.0) - loc
    assert distrib.draw((distrib.likely - loc) / scale) == pytest.approx(
        distrib.likely, abs=1e-7
    )

    # Worst and best inverted
    distrib = Triangular(8.0, 5.0, 3.0, pworst=0.3, pbest=0.05)

    assert distrib.draw(1.0 - distrib.pworst) == pytest.approx(distrib.worst, abs=1e-7)
    assert distrib.draw(distrib.pbest) == pytest.approx(distrib.best, abs=1e-7)

    loc = distrib.draw(0.0)
    scale = distrib.draw(1.0) - loc
    assert distrib.draw((distrib.likely - loc) / scale) == pytest.approx(
        distrib.likely, abs=1e-7
    )

    # Test without quantile
    distrib = Triangular(8.0, 9.0, 10.0, 0.0, 0.0)
    assert distrib.worst <= distrib.draw() <= distrib.best



import logging
import weakref
from enum import IntEnum
from functools import partial

try:
    from pythonfmu import Fmi2Status
except ImportError:

    class Fmi2Status(IntEnum):
        ok = 0
        warning = 1
        discard = 2
        error = 3
        fatal = 4


class FMUForwardHandler(logging.Handler):
    """Logging handler to forward Python log message to FMU logger.
    
    Attributes
    ----------
    _fmu : ReferenceType[pythonfmu.Fmi2Slave]
        FMU

    Parameters
    ----------
    fmu : pythonfmu.Fmi2Slave
        FMU to redirect log messages to
    """

    def __init__(self, fmu: "pythonfmu.Fmi2Slave", *args, **kwargs):
        super().__init__(*args, **kwargs)
        callback = partial(FMUForwardHandler.remove_handler, weakref.ref(self))
        self._fmu = weakref.ref(fmu, callback)

    @classmethod
    def add_handler(cls, fmu: "pythonfmu.Fmi2Slave", level: int) -> "FMUForwardHandler":
        """Add a new instance of this handler to the root logger.
        
        Parameters
        ----------
        fmu : pythonfmu.Fmi2Slave
            FMU

        Returns
        -------
        FMUForwardHandler
            The added handler
        """
        root_logger = logging.getLogger()
        if root_logger.level > level:
            root_logger.setLevel(level)
        handler = cls(fmu, level=level)
        root_logger.addHandler(handler)
        return handler

    @classmethod
    def remove_handler(
        cls, 
        handler: "ReferenceType[FMUForwardHandler]", 
        fmu: "Optional[ReferenceType[pythonfmu.Fmi2Slave]]" = None
    ) -> None:
        """Remove the instance of the provided handler on the
        root logger.
        
        Parameters
        ----------
        handler : weakref.ReferenceType[FMUForwardHandler]
            Weakreference object to the handler to be removed
        fmu : weakref.ReferenceType[pythonfmu.Fmi2Slave] or None
            Weakreference object to the FMU; default None
        """
        if handler() is not None:
            root_logger = logging.getLogger()
            root_logger.removeHandler(handler())

    @staticmethod
    def logging2FmiLevel(levelno: int) -> "pythonfmu.Fmi2Status":
        """Convert a Python log level to FMI 2 status.
        
        Parameters
        ----------
        levelno : int
            Python log level

        Returns
        -------
        pythonfmu.Fmi2Status
            FMI 2 status
        """
        if levelno >= logging.CRITICAL:
            return Fmi2Status.fatal
        elif levelno >= logging.ERROR:
            return Fmi2Status.error
        elif levelno >= logging.WARNING:
            return Fmi2Status.warning
        else:
            return Fmi2Status.ok

    def emit(self, record: "logging.LogRecord") -> None:
        """Forward a record to the FMU logger.
        
        Parameters
        ----------
        record : logging.LogRecord
            A log record
        """
        if self._fmu() is not None:
            self._fmu().log(
                record.getMessage(),
                FMUForwardHandler.logging2FmiLevel(record.levelno),
                debug=(record.levelno <= logging.DEBUG),
            )



import abc
from typing import Any, Dict, Iterable, Optional, Union

from cosapp.drivers.driver import Driver


# TODO
# [ ] Quid for vector variables
class AbstractSetOfCases(Driver):
    """
    This driver builds a set of cases from a list

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = ('cases')

    def __init__(
        self, name: str, owner: "Optional[cosapp.systems.System]" = None, **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)
        # TODO Fred - is this not too much oriented for MonteCarlo or DoE? What about a mission for which input
        # variables may not be the same on all points.
        self.cases = None  # type: Optional[Iterable[Any]]
            # desc="List of cases to be carried out."

    def _precase(self, case_idx: int, case: Any):
        """Hook to be called before running each case.
        
        Parameters
        ----------
        case_idx : int
            Index of the case
        case : Any
            Parameters for this case
        """
        self.status = ""
        self.error_code = "0"

    @abc.abstractmethod
    def _build_cases(self) -> None:
        """Generator of cases."""
        pass

    def _postcase(self, case_idx: int, case: Any):
        """Hook to be called before running each case.
        
        Parameters
        ----------
        case_idx : int
            Index of the case
        case : Any
            Parameters for this case
        """
        if self._recorder is not None:
            self._recorder.record_state(case_idx, self.status, self.error_code)

    def _precompute(self):
        """Actions performed prior to the `Module.compute` call."""
        super()._precompute()
        self._build_cases()

    def run_children(self) -> None:
        """Runs all driver children.
        """
        for child in self.children.values():
            child.run_once()
            if len(child.status) > 0:
                self.status = child.status
            if child.error_code != "0":
                self.error_code = child.error_code

    def compute(self) -> None:
        """Contains the customized `Module` calculation, to execute after children.
        """
        for case_idx, case in enumerate(self.cases):
            if len(case) > 0:
                self._precase(case_idx, case)
                self.run_children()
                self._postcase(case_idx, case)

    def run_once(self) -> None:
        """Run the system once.

        Execute the model of this `Module` and its children in the execution order.

        Notes
        -----

        The driver are not executed when calling this method; only the physical model.
        """
        if self.is_active():
            self._precompute()

            self.compute_before()
            self._compute_calls += 1
            self.compute()

            self._postcompute()
            self.computed.emit()



from pathlib import Path

import pytest

from cosapp.systems import System
from cosapp.drivers import NonLinearSolver


def test_system0(test_library, test_data):
    #
    #  pressureLoss0D
    #
    s = System.load(test_data / "system_config_pressureloss.json")
    s.run_once()
    assert s.flnum_out.Pt == 90000.0


def test_system1(test_library, test_data):
    #
    # pressureLossSys (pressureLoss0D)
    #
    s = System.load(test_data / "system_config_pressureloss1.json")
    s.exec_order = ["p1"]
    s.run_once()

    assert s.p1.flnum_out.Pt == 90000.0


def test_system2(test_library, test_data):
    #
    # pressureLossSys (pressureLoss0D + pressureLoss0D)
    #
    s = System.load(test_data / "system_config_pressureloss2.json")
    s.exec_order = ["p1", "p2"]
    s.run_once()

    assert s.p2.flnum_out.Pt == 80000.0


def test_system2tank(test_library, test_data):
    #
    # pressureLossSys (pressureLoss0D + tank + pressureLoss0D a pressureLossSys)
    #

    s = System.load(test_data / "system_config_pressureloss2tank.json")

    s.exec_order = ["p1", "tank1", "p2"]

    # tank is half full
    s.tank1.vol = 10.0
    s.run_once()
    assert s.p2.flnum_out.Pt == 89900.0
    assert s.tank1.vol == 10.9
    assert s.tank1.flnum_out.W == 1.0

    # tank is empty
    s.tank1.vol = 0.0
    s.run_once()
    assert s.p2.flnum_out.Pt == 89900.0
    assert s.tank1.vol == 0.9
    assert s.tank1.flnum_out.W == 1.0

    # tank is full
    s.tank1.vol = 100.0
    s.run_once()
    assert s.p2.flnum_out.Pt == 80000.0
    assert s.tank1.vol == pytest.approx(100)
    assert s.tank1.flnum_out.W == 10.0


def test_system11(test_library, test_data):
    #
    # pressureLossSys (pressureLossSys(pressureLoss0D))
    # with direct connexion of module
    #
    s = System.load(test_data / "system_config_pressureloss11.json")
    s.exec_order = ["p1"]
    s.p1.exec_order = ["p11"]
    s.run_once()

    assert s.p1.p11.flnum_out.Pt == pytest.approx(90000, rel=5e-8)


def test_system11bis(test_library, test_data):
    #
    # pressureLossSys (pressureLossSys(pressureLoss0D))
    # with connexion of module using ".."
    #
    s = System.load(test_data / "system_config_pressureloss11bis.json")
    s.run_once()

    assert s.p2.p21.flnum_out.Pt == pytest.approx(80000, rel=5e-8)
    assert s.p3.flnum_out.Pt == pytest.approx(70000, rel=5e-8)


def test_system12(test_library, test_data):
    #
    # pressureLossSys (pressureLossSys(pressureLoss0D + pressureLoss0D))
    #
    s = System.load(test_data / "system_config_pressureloss12.json")
    s.exec_order = ["p1"]
    s.p1.exec_order = ["p11", "p12"]
    s.run_once()

    assert s.p1.p12.flnum_out.Pt == pytest.approx(80000, rel=5.0e-8)


def test_system22(test_library, test_data):
    #
    # pressureLossSys (pressureLossSys(pressureLoss0D + pressureLoss0D) + pressureLoss0D)
    #
    s = System.load(test_data / "system_config_pressureloss22.json")
    s.exec_order = ["p1", "p2"]
    s.p1.exec_order = ["p11", "p12"]
    s.run_once()

    assert s.p2.flnum_out.Pt == pytest.approx(70000, rel=5.0e-8)


def test_system121(test_library, test_data):
    #
    # pressureLossSys (pressureLoss0D + Splitter12 + pressureLoss0D//pressureLoss0D + Mixer21 pressureLoss0D)
    #
    s = System.load(test_data / "system_config_pressureloss121.json")

    s.exec_order = ["p1", "sp", "p2", "p3", "mx", "p4"]
    s.run_once()

    assert s.p4.flnum_out.Pt == 74800.0

    d = s.add_driver(NonLinearSolver("design", tol=5.0e-8))
    d.runner.offdesign.add_unknown("sp.x")

    s.run_drivers()

    # test children of the system are listed in the execution order list
    assert len(s.children) == len(s.exec_order)
    for child in s.children:
        assert child in s.exec_order

    # test iterative loop detection
    assert len(s.residues) == 0
    residues = s.get_unsolved_problem().residues
    assert "mx.(epsilon == 0)" in residues

    assert s.p4.flnum_out.Pt == pytest.approx(77500, rel=1e-6)
    assert s.p2.flnum_out.W == pytest.approx(5, rel=1e-6)
    assert s.mx.flnum_out.W == pytest.approx(s.mx.flnum_in1.W + s.mx.flnum_in2.W, rel=1e-5)


def test_system131(test_library, test_data):
    #
    # pressureLossSys (pressureLoss0D + Mixer21 + pressureLoss0D//pressureLoss0D(backward) + Splitter12 + pressureLoss0D)
    #
    # ! Do not suppress this unit test, it's the only one doing new computation with resetting the system
    s = System.load(test_data / "system_config_pressureloss131.json")

    s.exec_order = ["p1", "mx", "p2", "sp", "p3", "p4"]

    s.p3.K = -12100
    s.p2.K = 100
    s.sp.x = 0.5

    d = s.add_driver(NonLinearSolver("design"))

    s.run_drivers()
    assert s.p4.flnum_out.Pt == pytest.approx(67900, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(11, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 11, rel=2e-5)

    s.p3.K = -3600
    s.run_drivers()
    print(d.problem)
    assert s.p4.flnum_out.Pt == pytest.approx(65600, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(12, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 12, rel=1e-5)

    s.p3.K = -12100
    d.options["factor"] = 0.7
    s.run_drivers()
    assert s.p4.flnum_out.Pt == pytest.approx(67900, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(11, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 11, rel=1e-5)

    s.p3.K = -260100
    # d.options["factor"] = 0.9
    s.run_drivers()

    assert s.p2.flnum_out.W == pytest.approx(10.2, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10.0, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 10.2, rel=1e-5)
    assert s.p4.flnum_out.Pt == pytest.approx(69596, rel=1e-5)


def test_system222(test_library, test_data):
    #
    # pLS(pl0 + mx + (pLS (pL0 + mx + (pL0+pL0//pl0(b) + sp + pl0))//pl0(b)) + sp + pL0)
    #
    s = System.load(test_data / "system_config_pressureloss222.json")

    s.exec_order = ["p1", "m1", "p2", "s1", "p3", "p4"]

    s.flnum_in.W = 9.0
    s.flnum_in.Pt = 108100
    s.p2.p22.p221.K = 50
    s.p2.p22.p222.K = 50
    s.p3.K = -32100
    s.p2.p23.K = -12100

    d = s.add_driver(NonLinearSolver("design"))
    d.runner.offdesign.add_unknown(["p2.s21.x", "s1.x"])
    s.run_drivers()

    assert s.p2.p24.flnum_out.Pt == pytest.approx(67900, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(9, rel=1e-5)
    assert s.s1.x == pytest.approx(9 / 10, rel=1e-5)
    assert s.p2.s21.x == pytest.approx(10 / 11, rel=1e-5)



import pytest

from cosapp.drivers.time.scenario import Interpolator, InterpolAssignString
from .conftest import ScalarOde


@pytest.fixture(scope='function')
def ode():
    return ScalarOde('ode')


@pytest.fixture(scope='function')
def data():
    return [[0, 0], [1, 1], [10, -17]]


def test_InterpolAssignString__init__(ode, data):
    interpolator = Interpolator(data)
    bc = InterpolAssignString('df', interpolator, ode)
    assert str(bc) == "df = Interpolator(t)"
    bc_locals = bc.locals
    assert len(bc_locals.keys()) == 3
    assert "t" in bc_locals
    assert "ode" in bc_locals
    assert bc_locals["ode"] is ode


def test_InterpolAssignString_error(ode, data):
    message = "Functions used in time boundary conditions may only be of type `Interpolator`"

    with pytest.raises(TypeError, match=message):
        InterpolAssignString('df', lambda t: 2 * t, ode)

    class CustomInterpolator(Interpolator):
        def __init__(self, data):
            super().__init__(data)
    
    with pytest.raises(TypeError, match=message):
        InterpolAssignString('df', CustomInterpolator(data), ode)


def test_InterpolAssignString_exec(ode, data, clock):
    interpolator = Interpolator(data)
    bc = InterpolAssignString('df', interpolator, ode)
    ode.df = 0.123
    clock.time = 0
    bc.exec()
    assert ode.df == 0
    clock.time = 0.5
    bc.exec()
    assert ode.df == pytest.approx(0.5, rel=1e-15)
    clock.time = 2
    bc.exec()
    assert ode.df == pytest.approx(-1, rel=1e-15)

    for clock.time in (0, 0.8, 1, 2, 3.4):
        bc.exec()
        assert ode.df == interpolator(clock.time)
        assert bc.locals['t'] == ode.time



"""Dummy call on D3 views"""
import pytest

from cosapp.tests.library.systems import FanComplex
from cosapp.tools.views.d3js import to_d3


pytest.importorskip("jinja2")


def test_to_d3():
    f = FanComplex('fan')
    to_d3(f, show=False)



from cosapp.systems import System

class ExpRampOde(System):
    """
    System representing function f(t) = a * (1 - exp(-t / tau)),
    through ODE: tau * f' + f = a
    
    f is then apply on an input signal x to produce y.
    """
    def setup(self):
        self.add_inward('a', 1.0)
        self.add_inward('tau', 1.0)
        self.add_inward('x', 1.0)
        self.add_outward('y', 1.0)

        self.add_outward('df_dt', 0.0)
        self.add_transient('f', der='df_dt', max_time_step='tau / 5')

    def compute(self):
        self.df_dt = (self.a - self.f) / self.tau
        self.y = self.x * self.f



from typing import List, Set, Union

import numpy
import pandas

from cosapp.drivers.abstractsetofcases import AbstractSetOfCases
from cosapp.drivers.optionaldriver import OptionalDriver
from cosapp.recorders.dataframe_recorder import DataFrameRecorder
from cosapp.utils.helpers import check_arg, is_number
from cosapp.utils.find_variables import find_variables


# TODO This does not support multipoints cases
class Influence(AbstractSetOfCases):
    """
    This driver evaluate the influence between inputs and outputs (floats only)
    In input :

    - the inwards (including in the ports)

    In output :

    - the inwards (including in the ports)
    - the connected ports
    - the outwards

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = (
        'input_vars', 'response_vars', 'influence_matrix', 'delta', 
        'influence_min_threshold', 'reference', 'found_input_vars', 'found_response_vars'
    )

    def __init__(
        self, name: str, owner: "Optional[cosapp.systems.System]" = None, **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)

        self.input_vars = ["*"]  # type: List[str]
            # desc="Variable names varying between cases"        
        self.response_vars = ["*"]  # type: List[str]
            # desc="Variable names to monitor between cases"
        self.influence_matrix = None  # type: Optional[numpy.ndarray]
            # desc="Influence matrix d/d output(d/d input)"
        self.delta = 1e-3  # type: float
            # desc="Relative influence to apply for influence matrix computation"
        self.influence_min_threshold = 1e-5  # type: float
            # desc="Minimum value of influence to consider"

        self.reference = pandas.DataFrame()  # type: pandas.DataFrame
        self.found_input_vars = list()  # type: List[str]
        self.found_response_vars = list()  # type: List[str]

    def reset_input_vars(self):
        """Reset input_vars variable to its default value ['*']
        """
        self.input_vars = ["*"]

    def reset_response_vars(self):
        """Reset response_vars variable to its default value ['*']
        """
        self.response_vars = ["*"]

    def add_input_vars(self, names: Union[str, Set[str], List[str]]):
        """Add variable to be influenced.

        Parameters
        ----------
        names : Union[str, Set[str], List[str]]
            List of variables to be influenced
        """
        self._add_vars(names, is_input=True)

    def add_response_vars(self, names: Union[str, Set[str], List[str]]):
        """Add variable to be monitored.

        Parameters
        ----------
        names : Union[str, Set[str], List[str]]
            List of variables to be monitored
        """
        self._add_vars(names, is_input=False)

    def _add_vars(self, names: Union[str, Set[str], List[str]], is_input: bool = True):
        """Add variable to be influenced or monitored.

        Parameters
        ----------
        names : Union[str, Set[str], List[str]]
            List of variables to be influenced or monitored
        is_input : bool
            Defines if the variables is an input or not
        """
        check_arg(names, 'names', (str, set, list))

        if is_input and self.input_vars and self.input_vars[0] == "*":
            self.input_vars.pop(0)
        if not is_input and self.response_vars and self.response_vars[0] == "*":
            self.response_vars.pop(0)

        if isinstance(names, str):
            if is_input:
                self.input_vars.append(names)
            self.response_vars.append(names)
        else:
            for name in names:                
                if isinstance(name, str):
                    if is_input:
                        self.input_vars.append(name)
                    self.response_vars.append(name)
                else:
                    raise TypeError(
                        f"string expected; got {type(name).__qualname__}."
                    )

    def _build_cases(self) -> None:
        """Build the list of cases to run during execution
        """
        self.add_recorder(
            DataFrameRecorder(
                includes=self.input_vars + self.response_vars,
                raw_output=True,
                numerical_only=True,
            )
        )

        self.found_input_vars = find_variables(
            self.owner,
            includes=self.input_vars,
            excludes=[],
            advanced_filter=lambda x: is_number(x),
            outputs=False,
        )
        self.found_response_vars = find_variables(
            self.owner,
            includes=self.response_vars,
            excludes=[],
            advanced_filter=lambda x: is_number(x),
            inputs=False,
        )

        # TODO('Filter on variable types to avoid bugs with strings etc')
        # TODO('Support sequences')
        def f(ref, idx) -> numpy.array:
            case = ref.copy()
            case[idx] *= 1 + self.delta
            return case

        case_ref = [self.owner[var] for var in self.found_input_vars]
        self.cases = [case_ref]
        self.cases.extend(
            [f(case_ref, index) for index in range(len(self.found_input_vars))]
        )

    def _precase(self, case_idx, case):
        """Hook to be called before running each case."""
        # TODO Use MonteCarlo machinery
        for variable, value in zip(self.found_input_vars, case):
            self.owner[variable] = value

    def _precompute(self):
        """Set execution order, build cases and run reference case."""
        super()._precompute()
        self._run_reference()
        OptionalDriver.set_inhibited(True)

    def _run_reference(self):
        """Run the reference case (i.e. with no influence), store the result and clean the recorder."""
        case = self.cases[0]
        if case:
            self._precase(0, case)
            self.run_children()
            self._postcase(0, case)
        self.cases.pop(0)

        data = self.recorder.export_data()
        self.reference = data.iloc[
            :, len(DataFrameRecorder.SPECIALS) :
        ]  # Skip the information
        self.recorder.start()

    def _postcompute(self):
        """Calculate the influence matrix"""
        OptionalDriver.set_inhibited(False)
        super()._postcompute()

        data = self.recorder.export_data()
        results = data.iloc[
            :, len(DataFrameRecorder.SPECIALS) :
        ]  # Skip the information

        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            deltas = pandas.DataFrame(
                results.values / self.reference.values - 1.0,
                columns=self.reference.columns,
            )

        # TODO Improve coverage of no influence cases and absolute vs relative influence or mix of influences
        deltas[numpy.abs(deltas) == numpy.inf] = 0
        deltas[numpy.abs(results.values) < 1e-4] = 0

        inputs = deltas[self.found_input_vars]
        inputs_diag = inputs.values.diagonal().reshape((-1, 1))
        outputs = deltas[self.found_response_vars]

        inputs_diag = numpy.where(inputs_diag == 0., 1e-15, inputs_diag)

        # find inputs modified during the children drivers exec (connected port, variable iterated, etc.)
        coeff = numpy.where(abs(inputs_diag - self.delta) < 1e-10, 1. / inputs_diag, 0.)

        self.influence_matrix = pandas.DataFrame(outputs.values * coeff,
                                                columns=outputs.columns,
                                                index=inputs.columns).fillna(0)

    def show_influence_matrix(self,
        cleaned: bool = True,
        styling: bool = False,
        transpose: bool = False,
        sort_by: str = None,
    ):
        """Return the influence matrix with cleaning and styling options.

        Parameters
        ----------
        cleaned : bool, optional
            Defines if the returned influence matrix will be cleaned or not
        styling : bool, optional
            Defines if the returned influence matrix will be styled or not
        transpose : bool, optional
            Defines if the matrix will transposed or not before return
        sort_by: str, optional
            If given, will sort
        """
        matrix = self.influence_matrix

        if cleaned:
            matrix = matrix.where((abs(matrix) > self.influence_min_threshold), 0.0)
            matrix = matrix[(matrix.T != 0).any()]
            matrix = matrix.loc[:, (matrix != 0).any()]

        if sort_by:
            matrix = matrix.reindex(
                matrix[sort_by].abs().sort_values(ascending=False).index
            )

        if transpose:
            matrix = matrix.T

        if styling:
            matrix = matrix.style.bar(align="mid", color=["#d65f5f", "#5fba7d"])

        return matrix



from cosapp.core.numerics.enum import NonLinearMethods

from cosapp.drivers.driver import Driver
from cosapp.drivers.runonce import RunOnce
from cosapp.drivers.runsinglecase import RunSingleCase
from cosapp.drivers.runoptim import RunOptim
from cosapp.drivers.nonlinearsolver import NonLinearSolver
from cosapp.drivers.optimizer import Optimizer
from cosapp.drivers.validitycheck import ValidityCheck
from cosapp.drivers.lineardoe import LinearDoE
from cosapp.drivers.montecarlo import MonteCarlo
from cosapp.drivers.metasystembuilder import MetaSystemBuilder
from cosapp.drivers.influence import Influence

from cosapp.drivers.time.euler import EulerExplicit
from cosapp.drivers.time.runge_kutta import RungeKutta

__all__ = [
    "Driver",
    "Influence",
    "LinearDoE",
    "NonLinearMethods",
    "NonLinearSolver",
    "MetaSystemBuilder",
    "MonteCarlo",
    "Optimizer",
    "RunOnce",
    "RunOptim",
    "RunSingleCase",
    "ValidityCheck",
    "EulerExplicit",
    "RungeKutta",
]
